from enum import Enum

from red_val.exceptions import RedSpecificationError


class InputType:
    class InputCategory(Enum):
        File = 0
        Directory = 1
        string = 2
        int = 3
        long = 4
        float = 5
        double = 6
        boolean = 7

    def __init__(self, input_category, is_array, is_optional):
        self.input_category = input_category
        self._is_array = is_array
        self._is_optional = is_optional

    @staticmethod
    def from_string(s):
        is_optional = s.endswith('?')
        if is_optional:
            s = s[:-1]

        is_array = s.endswith('[]')
        if is_array:
            s = s[:-2]

        input_category = None
        for ic in InputType.InputCategory:
            if s == ic.name:
                input_category = ic

        if input_category is None:
            raise RedSpecificationError('The given input type "{}" is not valid'.format(s))

        return InputType(input_category, is_array, is_optional)

    def to_string(self):
        return '{}{}{}'.format(self.input_category.name,
                               '[]' if self._is_array else '',
                               '?' if self._is_optional else '')

    def __repr__(self):
        return self.to_string()

    def __eq__(self, other):
        return (self.input_category == other.input_category) and \
               (self._is_array == other.is_array()) and \
               (self._is_optional == other.is_optional())

    def is_file(self):
        return self.input_category == InputType.InputCategory.File

    def is_directory(self):
        return self.input_category == InputType.InputCategory.Directory

    def is_array(self):
        return self._is_array

    def is_optional(self):
        return self._is_optional

    def is_primitive(self):
        return (self.input_category != InputType.InputCategory.Directory) and \
               (self.input_category != InputType.InputCategory.File)


class OutputType:
    class OutputCategory(Enum):
        File = 0
        Directory = 1
        stdout = 2
        stderr = 3

    def __init__(self, output_category, is_optional):
        self.output_category = output_category
        self._is_optional = is_optional

    @staticmethod
    def from_string(s):
        is_optional = s.endswith('?')
        if is_optional:
            s = s[:-1]

        output_category = None
        for oc in OutputType.OutputCategory:
            if s == oc.name:
                output_category = oc

        if output_category is None:
            raise RedSpecificationError('The given output type "{}" is not valid'.format(s))

        if output_category == OutputType.OutputCategory.stdout and is_optional:
            raise RedSpecificationError(
                'The given output type is an optional stdout ("{}"), which is not valid'.format(s)
            )
        if output_category == OutputType.OutputCategory.stderr and is_optional:
            raise RedSpecificationError(
                'The given output type is an optional stderr ("{}"), which is not valid'.format(s)
            )

        return OutputType(output_category, is_optional)

    def to_string(self):
        return '{}{}'.format(
            self.output_category.name,
            '?' if self._is_optional else ''
        )

    def __repr__(self):
        return self.to_string()

    def __eq__(self, other):
        return (self.output_category == other.output_category) and \
               (self._is_optional == other.is_optional())

    # noinspection PyMethodMayBeStatic
    def is_array(self):
        return False

    def is_file(self):
        return self.output_category == OutputType.OutputCategory.File

    def is_directory(self):
        return self.output_category == OutputType.OutputCategory.Directory

    def is_stdout(self):
        return self.output_category == OutputType.OutputCategory.stdout

    def is_stderr(self):
        return self.output_category == OutputType.OutputCategory.stderr

    def is_stream(self):
        """
        Returns True, if this OutputType holds a stdout or stderr
        """
        return self.is_stdout() or self.is_stderr()

    def is_optional(self):
        return self._is_optional



from red_val.schemas.cli import cli_schema, PATTERN_KEY
from red_val.schemas.connector import connector_schema
from red_val.schemas.engine import engine_schema

red_schema = {
    'definitions': {
        'redVersion': {'enum': ['9']},
        'cli': cli_schema,
        'connector': connector_schema,
        'engine': engine_schema,
        'listingFile': {
            'type': 'object',
            'properties': {
                'class': {'enum': ['File']},
                'basename': {'type': 'string'},
                'checksum': {'type': 'string'},
                'size': {'type': 'integer'}
            },
            'required': ['class', 'basename'],
            'additionalProperties': False
        },
        'listingDirectory': {
            'type': 'object',
            'properties': {
                'class': {'enum': ['Directory']},
                'basename': {'type': 'string'},
                'listing': {'$ref': '#/definitions/listing'}
            },
            'additionalProperties': False,
            'required': ['class', 'basename']
        },
        'listing': {
            'type': 'array',
            'items': {
                'oneOf': [{'$ref': '#/definitions/listingFile'}, {'$ref': '#/definitions/listingDirectory'}]
            }
        },
        'inputFile': {
            'type': 'object',
            'properties': {
                'class': {'enum': ['File']},
                'connector': {'$ref': '#/definitions/connector'},
                'basename': {'type': 'string'},
                'dirname': {'type': 'string'},
                'checksum': {'type': 'string'},
                'size': {'type': 'integer'},
                'doc': {'type': 'string'}
            },
            'additionalProperties': False,
            'required': ['class', 'connector']
        },
        'inputDirectory': {
            'type': 'object',
            'properties': {
                'class': {'enum': ['Directory']},
                'connector': {'$ref': '#/definitions/connector'},
                'basename': {'type': 'string'},
                'listing': {'$ref': '#/definitions/listing'},
                'doc': {'type': 'string'}
            },
            'additionalProperties': False,
            'required': ['class', 'connector']
        },
        'outputFile': {
            'type': 'object',
            'properties': {
                'class': {'enum': ['File', 'stdout', 'stderr']},
                'checksum': {'type': 'string'},
                'size': {'type': 'integer'},
                'connector': {'$ref': '#/definitions/connector'},
                'doc': {'type': 'string'}
            },
            'additionalProperties': False,
            'required': ['class', 'connector']
        },
        'outputDirectory': {
            'type': 'object',
            'properties': {
                'class': {'enum': ['Directory']},
                'connector': {'$ref': '#/definitions/connector'},
                'listing': {'$ref': '#/definitions/listing'},
                'doc': {'type': 'string'}
            },
            'additionalProperties': False,
            'required': ['class', 'connector']
        },
        'inputs': {
            'type': 'object',
            'patternProperties': {
                PATTERN_KEY: {
                    'anyOf': [
                        {'type': 'string'},
                        {'type': 'number'},
                        {'type': 'boolean'},
                        {
                            'type': 'array',
                            'items': {
                                'oneOf': [
                                    {'type': 'string'},
                                    {'type': 'number'},
                                    {'type': 'boolean'},
                                    {'$ref': '#/definitions/inputFile'},
                                    {'$ref': '#/definitions/inputDirectory'}
                                ]
                            }
                        },
                        {'$ref': '#/definitions/inputFile'},
                        {'$ref': '#/definitions/inputDirectory'}
                    ]
                }
            },
            'additionalProperties': False
        },
        'outputs': {
            'type': 'object',
            'patternProperties': {
                PATTERN_KEY: {
                    'anyOf': [
                        {'$ref': '#/definitions/outputFile'},
                        {'$ref': '#/definitions/outputDirectory'}
                    ]
                }
            },
            'additionalProperties': False
        }
    },
    'oneOf': [{
        'type': 'object',
        'properties': {
            'redVersion': {'$ref': '#/definitions/redVersion'},
            'cli': {'$ref': '#/definitions/cli'},
            'inputs': {'$ref': '#/definitions/inputs'},
            'outputs': {'$ref': '#/definitions/outputs'},
            'container': {'$ref': '#/definitions/engine'},
            'execution': {'$ref': '#/definitions/engine'},
            'doc': {'type': 'string'}
        },
        'additionalProperties': False,
        'required': ['redVersion', 'cli', 'inputs', 'container']
    }, {
        'type': 'object',
        'properties': {
            'redVersion': {'$ref': '#/definitions/redVersion'},
            'cli': {'$ref': '#/definitions/cli'},
            'batches': {
                'type': 'array',
                'items': {
                    'type': 'object',
                    'properties': {
                        'inputs': {'$ref': '#/definitions/inputs'},
                        'outputs': {'$ref': '#/definitions/outputs'},
                        'doc': {'type': 'string'}
                    },
                    'additionalProperties': False,
                    'required': ['inputs']
                }
            },
            'container': {'$ref': '#/definitions/engine'},
            'execution': {'$ref': '#/definitions/engine'},
            'doc': {'type': 'string'}
        },
        'additionalProperties': False,
        'required': ['redVersion', 'cli', 'batches', 'container']
    }]
}



URL_SCHEME_IDENTIFIER = 'path'

CWL_INPUT_TYPES = ['File', 'Directory', 'string', 'int', 'long', 'float', 'double', 'boolean']
CWL_INPUT_TYPES += ['{}[]'.format(t) for t in CWL_INPUT_TYPES[:]]
CWL_INPUT_TYPES += ['{}?'.format(t) for t in CWL_INPUT_TYPES[:]]

CWL_OUTPUT_TYPES = ['File', 'Directory']
CWL_OUTPUT_TYPES += ['{}?'.format(t) for t in CWL_OUTPUT_TYPES[:]]

PATTERN_KEY = '^[a-zA-Z0-9_-]+$'


cli_schema = {
    'type': 'object',
    'properties': {
        'cwlVersion': {'enum': ['v1.0']},
        'class': {'enum': ['CommandLineTool']},
        'baseCommand': {
            'oneOf': [
                {'type': 'string'},
                {'type': 'null'},
                {
                    'type': 'array',
                    'items': {'type': 'string'}
                }
            ]
        },
        'inputs': {
            'type': 'object',
            'patternProperties': {
                PATTERN_KEY: {
                    'type': 'object',
                    'properties': {
                        'type': {'enum': CWL_INPUT_TYPES},
                        'inputBinding': {
                            'type': 'object',
                            'properties': {
                                'prefix': {'type': 'string'},
                                'separate': {'type': 'boolean'},
                                'position': {'type': 'integer', 'minimum': 0},
                                'itemSeparator': {'type': 'string'}
                            },
                            'additionalProperties': False,
                        },
                        'doc': {'type': 'string'}
                    },
                    'additionalProperties': False,
                    'required': ['type', 'inputBinding']
                }
            }
        },
        'outputs': {
            'type': 'object',
            'patternProperties': {
                PATTERN_KEY: {
                    'oneOf': [{
                        'type': 'object',
                        'properties': {
                            'type': {'enum': CWL_OUTPUT_TYPES},
                            'outputBinding': {
                                'type': 'object',
                                'properties': {
                                    'glob': {'type': 'string'},
                                },
                                'additionalProperties': False,
                                'required': ['glob']
                            },
                            'doc': {'type': 'string'}
                        },
                        'additionalProperties': False,
                        'required': ['type', 'outputBinding']
                    }, {
                        'type': 'object',
                        'properties': {
                            'type': {'enum': ['stdout', 'stderr']},
                        },
                        'additionalProperties': False,
                        'required': ['type']
                    }]
                }
            }
        },
        'stdout': {'type': 'string'},
        'stderr': {'type': 'string'},
        'doc': {'type': 'string'}
    },
    'additionalProperties': False,
    'required': ['cwlVersion', 'class', 'inputs', 'outputs']
}



def _partition_all_internal(s, sep):
    """
    Uses str.partition() to split every occurrence of sep in s. The returned list does not contain empty strings.

    :param s: The string to split.
    :param sep: A separator string.
    :return: A list of parts split by sep
    """
    parts = list(s.partition(sep))

    # if sep found
    if parts[1] == sep:
        new_parts = partition_all(parts[2], sep)
        parts.pop()
        parts.extend(new_parts)
        return [p for p in parts if p]
    else:
        if parts[0]:
            return [parts[0]]
        else:
            return []


def partition_all(s, sep):
    """
    Uses str.partition() to split every occurrence of sep in s. The returned list does not contain empty strings.
    If sep is a list, all separators are evaluated.

    :param s: The string to split.
    :param sep: A separator string or a list of separator strings.
    :return: A list of parts split by sep
    """
    if isinstance(sep, list):
        parts = _partition_all_internal(s, sep[0])
        sep = sep[1:]

        for s in sep:
            tmp = []
            for p in parts:
                tmp.extend(_partition_all_internal(p, s))
            parts = tmp

        return parts
    else:
        return _partition_all_internal(s, sep)


def split_all(reference, sep):
    """
    Splits a given string at a given separator or list of separators.

    :param reference: The reference to split.
    :param sep: Separator string or list of separator strings.
    :return: A list of split strings
    """
    parts = partition_all(reference, sep)
    return [p for p in parts if p not in sep]


def strip_start_end(parts, start_symbol, end_symbol):
    """
    For every part in parts remove the start_symbol from the begin of part and the end_symbol from the end of part if
    both are present. Otherwise (if start_symbol or end_symbol is missing) leaves part unchanged.
    :param parts: A list of strings, which may start with start_symbol and end with end_symbol
    :type parts: List[str]
    :param start_symbol: The symbol to remove from the start of every part, if end_symbol is present at the end of this
    part.
    :param end_symbol: The symbol to remove from the end of every part, if start_symbol is present at the start of this
    part.
    :return: A new list containing the given parts but with removed start_symbol and end_symbol
    """
    stripped_parts = []

    for part in parts:
        if part.startswith(start_symbol) and part.endswith(end_symbol):
            stripped_part = part.lstrip(start_symbol).rstrip(end_symbol)
            stripped_parts.append(stripped_part)
        else:
            stripped_parts.append(part)

    return stripped_parts


def split_into_parts_with_separators(to_split, separator_list, remove_separators=False):
    """
    Splits the given string into parts, which either start with one of the start separators and end with the
    corresponding end separator or do not contain separators.
    If remove_separators is set to True, the parts are split in the same way, but afterwards all separators are removed
    from the resulting parts.

    :param to_split: The string to split
    :param separator_list: A list containing tuples of separator starts and ends.
    Example: [('start', 'end'), ('<', '>'), ...]
    :param remove_separators: Defines whether the separator symbols should be removed from the output.
    :raise ParsingError: If separator start is found, but the corresponding end separator does not occur until the end
    of to_split
    :return: The given string split into parts, by separators given in separator list.
    """
    # parts is a list containing strings, which will be split by the following procedure
    parts = [to_split]
    for start_separator, end_separator in separator_list:
        tmp_parts = []
        for p in parts:
            tmp_parts.extend(split_into_parts(p, start_separator, end_separator, remove_separators))
        parts = tmp_parts

    return parts


def split_into_parts(to_split, start, end, remove_separators=False):
    """
    Splits to_split in normal strings and strings, that start with the given start separator and end with the given end
    separator.
    If remove_separators is set to True, to_split is split in the same way, but separators are removed afterwards.

    Example:
    split_into_parts("a(b)cde()(fg)", start='(', end=')') == ["a", "(b)", "cde", "()", "(fg)"]

    :param to_split: The string to split
    :param start: The start sequence to search for
    :param end: The end sequence to search for
    :param remove_separators: If set to True, separators are removed from the output parts
    :raise ParsingError: If an input reference is not closed and a new reference starts or the string ends.
    :return: A list of normal strings and unresolved input references.
    """
    parts = partition_all(to_split, [start, end])

    result = []
    part = []
    in_reference = False
    for p in parts:
        if in_reference:
            if p == start:
                raise ParsingError('A new framed string has been started, although the old framed string has not yet '
                                   'been completed.\n{}'.format(to_split))
            elif p == end:
                part.append(end)
                result.append(''.join(part))
                part = []
                in_reference = False
            else:
                part.append(p)
        else:
            if p == start:
                if part:
                    result.append(''.join(part))
                part = [start]
                in_reference = True
            else:
                part.append(p)

    if in_reference:
        raise ParsingError(
            'string "{}" contains start symbol "{}" but not end symbol "{}"'.format(to_split, start, end)
        )
    elif part:
        result.append(''.join(part))

    if remove_separators:
        result = strip_start_end(result, start, end)

    return result


class ParsingError(Exception):
    pass



import itertools
import os

import jsonschema
from jsonschema import ValidationError
from red_val.exceptions import RedSpecificationError, RedValidationError, CWLSpecificationError, format_validation_error
from red_val.schemas.red import red_schema
from red_val.red_types import InputType, OutputType
from red_val.red_variables import get_variable_keys, RedVariableError


def red_validation(red_data, ignore_outputs, container_requirement=False, allow_variables=True):
    """
    Checks the given red data. The process implements the following steps:

    - check if all keys in the red data are strings
    - validate red data with schema
    - validate if RED version of the RED file matches the version of red-val
    - check if all given job values have a corresponding cli description
    - match types of cli description and job values
    - validate listings given in connectors
    - validate container requirement
    - check whether template keys are present, if they are not allowed
    - checks if globs do start with "/"

    Not checked anymore:
    - directory listings

    :param red_data: The red data to check
    :param ignore_outputs: Whether the outputs section should be ignored
    :param container_requirement: If True, this function checks, if there is a container section in the red file
    :param allow_variables: Whether red variables are allowed in the given red data or not
    :raise RedValidationError, RedSpecificationError, CWLSpecificationError: If the red data is not valid
    """
    check_keys_are_strings(red_data)

    try:
        jsonschema.validate(red_data, red_schema)
    except ValidationError as e:
        raise RedValidationError(
            format_validation_error(e, 'RED-FILE does not comply with jsonschema:')
        )

    input_cli_job_pairs, output_cli_job_pairs = _create_cli_job_pairs(red_data, ignore_outputs)

    # check whether types of job data do fit to cli description
    for cli_job_pair in itertools.chain(input_cli_job_pairs, output_cli_job_pairs):
        cli_job_pair.check_type()

        # directory listings are not checked
        # cli_job_pair.check_directory_listing()

    # check for forbidden variables
    if not allow_variables:
        variables = get_variable_keys(red_data)
        if variables:
            raise RedVariableError(
                'Red variables are forbidden, but found the following: {}'.format([v.key for v in variables])
            )

    if container_requirement:
        if not red_data.get('container'):
            raise RedSpecificationError('container engine description is missing in REDFILE')

    _check_output_glob(red_data)


def _check_output_glob(red_data):
    """
    Raises an CwlSpecificationError, if a glob is given as absolute path.

    :param red_data: The red data to analyse
    """
    cli_outputs = red_data['cli'].get('outputs')
    if cli_outputs:
        for output_key, output_value in cli_outputs.items():
            if output_value['type'] == 'stdout' or output_value['type'] == 'stderr':
                continue
            glob = output_value['outputBinding']['glob']
            if os.path.isabs(glob):
                raise CWLSpecificationError(
                    'Glob of output key "{}" starts with "/", which is illegal'.format(output_key)
                )


class CliJobPair:
    """
    A CliJobPair represents a cli description of one input/output key and a corresponding input/output batch value.
    """
    def __init__(self, key, is_input, cli_description, job_value):
        """
        Initializes a CliJobPair
        :param key: The input/output key of the cli description and job value
        :param is_input: Defines whether it is an input or an output key
        :param cli_description: A dictionary containing a cli description of an input/output key.
        Keys of this cli description are ['type', 'inputBinding']
        :param job_value: A primitive type or a dictionary containing a job value of an input or output
        file/directory. In case of a file/directory the keys of the dictionary are ['class', 'connector']
        """
        self.key = key
        self.is_input = is_input
        self.cli_description = cli_description
        self.job_value = job_value

    def _get_input_output(self):
        return 'input' if self.is_input else 'output'

    def check_type(self):
        """
        Checks whether the job value type fits to the cli description

        :raises RedSpecificationError: if cli_description and job value have incompatible types
        """
        try:
            if self.is_input:
                _check_input_type(self.job_value, self.cli_description['type'])
            else:
                _check_output_type(self.job_value, self.cli_description['type'])
        except RedSpecificationError as e:
            raise RedSpecificationError(
                'Error while checking {} key "{}":\n{}'.format(self._get_input_output(), self.key, str(e))
            )

    def check_directory_listing(self):
        """
        Validates a possible directory listing

        :raise RedValidationError: If listing does not match given job data
        """
        # if self.is_input:
        #     cli_type = InputType.from_string(self.cli_description['type'])
        # else:
        #     cli_type = OutputType.from_string(self.cli_description['type'])

        # # handle job value always as array
        # if cli_type.is_array():
        #     job_values = self.job_value
        # else:
        #     job_values = [self.job_value]

        # for job_value in job_values:
        #     if cli_type.is_directory() and job_value is not None:
        #         try:
        #             _red_listing_validation(job_value.get('listing'))
        #         except RedValidationError as e:
        #             raise RedValidationError(
        #                 'Error while checking listing for {} key "{}":\n{}'
        #                 .format(self._get_input_output(), self.key, str(e))
        #             )


# def _red_listing_validation(listing):
#     """
#     Raises an RedValidationError, if the given listing does not comply with cwl_job_listing_schema.
#     If listing is None or an empty list, no exception is thrown.
#
#     :param listing: The listing to validate
#     :raise RedValidationError: If the given listing does not comply with cwl_job_listing_schema
#     """
#
#     if listing:
#         try:
#             jsonschema.validate(listing, cwl_job_listing_schema)
#         except ValidationError as e:
#             where = '.'.join([str(s) for s in e.absolute_path]) if e.absolute_path else '/'
#             raise RedValidationError(
#                 'listing does not comply with jsonschema:\n\t key: {}\n\t reason: {}'.format(where, e.message)
#             )


def _create_cli_job_pairs(red_data, ignore_outputs):
    """
    Creates one list for all input cli_job pairs and one for all output cli_job pairs given in red data and returns them
    as tuple. If ignore_outputs is True, the output cli_job pair list is empty.

    :param red_data: The red data to get cli job pairs from
    :param ignore_outputs: Whether to ignore outputs or not
    :return: A tuple(input_cli_job_pairs, output_cli_job_pairs) with:
             - input_cli_job_pairs: A list of all CliJobPairs given in the input section of red_data
             - output_cli_job_pairs: A list of all CliJobPairs given in the output section of red_data
    :rtype: Tuple[List[CliJobPair], List[CliJobPair]]
    :raise RedSpecificationError: If there is a job value, but no corresponding cli description
    """
    batches = red_data.get('batches')
    if batches is None:
        batches = [{
            'inputs': red_data['inputs'],
            'outputs': red_data.get('outputs')
        }]

    input_cli_job_pairs = []

    # input cli job pairs
    cli_inputs = red_data['cli']['inputs']
    for batch in batches:
        job_inputs = batch['inputs']
        input_keys = set.union(set(batch['inputs'].keys()), set(cli_inputs.keys()))

        for input_key in input_keys:
            if input_key not in cli_inputs:
                raise RedSpecificationError(
                    'Input key "{}" is used in job description, but is not given in cli description'.format(input_key)
                )

            input_cli_job_pair = CliJobPair(input_key, True, cli_inputs[input_key], job_inputs.get(input_key))
            input_cli_job_pairs.append(input_cli_job_pair)

    output_cli_job_pairs = []

    # output cli job pairs
    if not ignore_outputs:
        cli_outputs = red_data['cli']['outputs']

        for batch in batches:
            job_outputs = batch.get('outputs')
            if job_outputs is None:
                continue
            output_keys = set.union(set(batch['outputs'].keys()), set(cli_outputs.keys()))

            for output_key in output_keys:
                if output_key not in cli_outputs:
                    raise RedSpecificationError(
                        'Output key "{}" is used in job description, but is not given in cli description'
                        .format(output_key)
                    )

                output_cli_job_pair = CliJobPair(
                    output_key,
                    False,
                    cli_outputs[output_key],
                    job_outputs.get(output_key)
                )
                output_cli_job_pairs.append(output_cli_job_pair)

    return input_cli_job_pairs, output_cli_job_pairs


CWL_INPUT_TYPE_TO_PYTHON_TYPE = {
    InputType.InputCategory.File: {dict},
    InputType.InputCategory.Directory: {dict},
    InputType.InputCategory.string: {str},
    InputType.InputCategory.int: {int},
    InputType.InputCategory.long: {int},
    InputType.InputCategory.float: {float, int},
    InputType.InputCategory.double: {float, int},
    InputType.InputCategory.boolean: {bool},
}

CWL_OUTPUT_TYPE_TO_PYTHON_TYPE = {
    OutputType.OutputCategory.File: {dict},
    OutputType.OutputCategory.Directory: {dict},
    OutputType.OutputCategory.stdout: {dict},
    OutputType.OutputCategory.stderr: {dict}
}


def _check_input_type(input_value, cli_description_type):
    """
    Checks whether the type of the given input value matches the type of the given cli description.

    :param input_value: The input value whose type to check
    :param cli_description_type: The cwl type description of the input key
    :raise RedSpecificationError: If actual input type does not match type of cli description
    """
    input_type = InputType.from_string(cli_description_type)

    # check whether job value is specified
    if input_value is None:
        if input_type.is_optional():
            return
        raise RedSpecificationError('Input value specification is missing and not optional.')

    # check for arrays
    # after this block input_value is always an array and an error is thrown, if input value has wrong list type
    if input_type.is_array():
        if not isinstance(input_value, list):
            raise RedSpecificationError('cli is declared as array, but value is not given as such')
    else:
        if isinstance(input_value, list):
            raise RedSpecificationError('cli is not declared as array, but value is given as array')
        input_value = [input_value]

    for sub_input_value in input_value:
        set_of_possible_value_types = CWL_INPUT_TYPE_TO_PYTHON_TYPE[input_type.input_category]
        if type(sub_input_value) not in set_of_possible_value_types:
            if isinstance(sub_input_value, dict):
                short_repr = 'dictionary'
            elif isinstance(sub_input_value, list):
                short_repr = 'list'
            else:
                short_repr = 'value "{}" of type "{}"'.format(sub_input_value, type(sub_input_value).__name__)

            raise RedSpecificationError('Value should have type "{}", but found "{}".'.format(
                input_type.input_category.name, short_repr
            ))

        if not input_type.is_primitive():
            cli_type = input_type.input_category.name
            value_type = sub_input_value.get('class')
            if cli_type != value_type:
                raise RedSpecificationError('Is declared as "{}" but given as "{}"'.format(
                    cli_type, value_type
                ))


def _check_output_type(output_value, cli_description_type):
    """
    Checks whether the type of the given output value matches the type of the given cli description.
    :param output_value: The output value whose type to check
    :param cli_description_type: The cwl type description of the output key

    :raise RedSpecificationError: If actual output type does not match type of cli description
    """
    output_type = OutputType.from_string(cli_description_type)

    # check whether the job value is specified
    if output_value is None:
        return

    set_of_possible_value_types = CWL_OUTPUT_TYPE_TO_PYTHON_TYPE[output_type.output_category]
    if type(output_value) not in set_of_possible_value_types:
        if isinstance(output_value, dict):
            short_repr = 'dictionary'
        elif isinstance(output_value, list):
            short_repr = 'list'
        else:
            short_repr = 'value "{}" of type "{}"'.format(output_value, type(output_value).__name__)

        raise RedSpecificationError('Value should have type "{}", but found "{}".'.format(
            output_type.output_category.name, short_repr
        ))

    cli_type = output_type.output_category.name
    value_type = output_value.get('class')
    if cli_type != value_type:
        raise RedSpecificationError('Is declared as "{}" but given as "{}"'.format(
            cli_type, value_type
        ))


def check_keys_are_strings(data, path=None):
    """
    Raises an RedSpecificationError, if a key is not of type string

    :param data: The data to check
    :param path: The path of keys as list of strings leading to data
    :raise RedSpecificationError: If a key is found, that has a type different from str
    """
    if path is None:
        path = []

    if isinstance(data, dict):
        for key, value in data.items():
            _check_key_is_string(key, path)
            check_keys_are_strings(value, path + [key])
    elif isinstance(data, list):
        for index, value in enumerate(data):
            check_keys_are_strings(value, path + [str(index)])


def _check_key_is_string(key, path):
    """
    Raises an RedSpecificationError, if the given key is not of type string.

    :param key: The key to check the type
    :param path: The path to this key
    :raise RedSpecificationError: If the given key has a type different from str
    """
    if not isinstance(key, str):
        if path:
            where = 'under "{}" '.format('.'.join(path))
        else:
            where = ''
        raise RedSpecificationError(
            'The key "{}" ({}) in REDFILE {}is not of type string'.format(key, type(key).__name__, where)
        )



from red_val.parsing import split_into_parts, ParsingError


VARIABLE_SEPARATOR_START = '{{'
VARIABLE_SEPARATOR_END = '}}'

PRIVATE_KEYS = {'access', 'auth'}
DEFAULT_PROTECTED_KEYS = {'password', 'privateKey'}


class VariableKey:
    def __init__(self, key, protected):
        """
        Creates a new VariableKey

        :param key: The key string of this VariableKey
        :param protected: Indicates if this VariableKey is protected or not
        """
        self.key = key
        self.protected = protected

    def __str__(self):
        return self.key


def get_variable_keys(data):
    """
    Returns a list of variable keys found in the given data.
    A variable key string is a string that starts with '{{' and ends with '}}'.
    To collect the variable keys iterates recursively over data values and appends variable keys to the variable keys
    set.

    :param data: The data to analyse.
    :type data: dict
    :return: A list of unique template keys
    """
    variable_keys = set()
    get_variable_keys_impl(data, variable_keys)
    return unique_variable_keys(variable_keys)


def get_variable_keys_impl(data, variable_keys, key_string=None, variable_keys_allowed=False, protected=False):
    """
    Iterates recursively over data values and appends variable keys to the variable keys list.
    A variable key string is a string that starts with '{{' and ends with '}}'.

    :param data: The data to analyse.
    :param variable_keys: A set of variable to append variable keys to.
    :type variable_keys: set
    :param key_string: A string representing the keys above the current data element.
    :param variable_keys_allowed: A boolean that specifies whether variable keys are allowed in the current dict
                                  position. If a variable key is found but it is not allowed an exception is thrown.
    :param protected: Indicates that the sub keys should be treated as protected keys

    :raise RedVariableError: If a variable key is found, but is not allowed.
    """
    if isinstance(data, dict):
        for key, value in data.items():
            sub_key_string = get_dict_sub_key_string(key, key_string)

            sub_variable_keys_allowed = variable_keys_allowed or (key in PRIVATE_KEYS)
            sub_protected = protected or is_protected_key(key)

            if sub_protected and not sub_variable_keys_allowed:
                raise RedVariableError(
                    'Found protected key "{}", but protected keys are only allowed under one of {}'
                    .format(sub_key_string, str(PRIVATE_KEYS))
                )
            get_variable_keys_impl(
                data=value,
                variable_keys=variable_keys,
                key_string=sub_key_string,
                variable_keys_allowed=sub_variable_keys_allowed,
                protected=sub_protected
            )
    elif isinstance(data, list):
        for index, sub_data in enumerate(data):
            sub_key_string = get_list_sub_key_string(index, key_string)

            get_variable_keys_impl(
                data=sub_data,
                variable_keys=variable_keys,
                key_string=sub_key_string,
                variable_keys_allowed=variable_keys_allowed,
                protected=protected
            )
    elif isinstance(data, str):
        if variable_keys_allowed:
            new_variable_keys = _extract_variable_keys(data, key_string, protected)
            if new_variable_keys:
                variable_keys.update(new_variable_keys)
        elif (VARIABLE_SEPARATOR_START in data) or (VARIABLE_SEPARATOR_END in data):
            raise RedVariableError(
                'Found invalid bracket in "{}" under "{}" in red data. Variable keys are only allowed as sub element of'
                ' an auth or access key.'.format(data, key_string)
            )


def unique_variable_keys(variable_keys):
    """
    Returns a sorted list of variable keys, where every key is unique.

    :param variable_keys: The list of variable keys from which duplicates are to be filtered out
    :type variable_keys: set[VariableKey]
    :return: A list of unique variable keys
    :rtype: list[VariableKey]
    """
    d = {}
    for variable_key in variable_keys:
        if variable_key.protected:
            d[variable_key.key] = variable_key
        else:
            if variable_key.key not in d:
                d[variable_key.key] = variable_key

    variable_key_list = list(d.values())
    variable_key_list.sort(key=lambda key: key.key, reverse=True)
    return variable_key_list


def get_dict_sub_key_string(key, key_string):
    """
    Helper function to add the given key to preceding keys.
    Handles the case, if key is the first key and key_string is None.

    :param key: The key to add to the preceding keys
    :param key_string: The string containing the preceding keys
    :return: A string concatenation of key_string and key
    """
    if key_string is None:
        sub_key_string = key
    else:
        sub_key_string = '{}.{}'.format(key_string, key)
    return sub_key_string


def get_list_sub_key_string(index, key_string):
    """
    Helper function to add the given index to preceding keys.
    Handles the case, if key is the first key and key_string is None.

    :param index: The index to add to the key_string
    :param key_string: The string containing the preceding keys
    :return: A string concatenation of key_string and index
    """
    if key_string is None:
        sub_key_string = '[{}]'.format(index)
    else:
        sub_key_string = '{}[{}]'.format(key_string, index)
    return sub_key_string


def is_protected_key(key):
    """
    Returns whether the given key is a protected key. (a key that is protected by default or starts with an underscore).

    :param key: The key to check
    :return: True, if the given key is a protected key, otherwise False
    """
    return (key in DEFAULT_PROTECTED_KEYS) or (key.startswith('_'))


def _extract_variable_keys(variable_string, key_string, protected):
    """
    Returns a set of variable keys, found inside variable_string

    :param variable_string: The string to analyse
    :type variable_string: str
    :param key_string: The keys of the given variable_string
    :param protected: Indicates whether the extracted keys are protected
    :return: A set of variable keys found inside variable_string

    :raise Parsing: If the variable_string is malformed
    """
    try:
        parts = split_into_parts(variable_string, VARIABLE_SEPARATOR_START, VARIABLE_SEPARATOR_END)
    except ParsingError as e:
        raise RedVariableError(
            'Could not parse variables from string "{}" in "{}". Do not use "{{" or "}}" in strings except for '
            'variable values. Failed with the following message:\n{}'.format(variable_string, key_string, str(e))
        )

    variable_keys = set()

    for part in parts:
        if is_variable_key(part):
            variable_key_string = part[2:-2]
            if (VARIABLE_SEPARATOR_START in variable_key_string) or (VARIABLE_SEPARATOR_END in variable_key_string):
                raise RedVariableError(
                    'Could not parse variable string "{}" in "{}". Too many brackets.'
                    .format(variable_string, key_string)
                )
            if variable_key_string == '':
                raise RedVariableError(
                    'Could not parse variable string "{}" in "{}". Variable keys should not be empty.'
                    .format(variable_string, key_string)
                )
            variable_keys.add(VariableKey(variable_key_string, protected))
        elif (VARIABLE_SEPARATOR_START in part) or (VARIABLE_SEPARATOR_END in part):
            raise RedVariableError(
                'Could not parse variable string "{}" in "{}". Too many brackets.'.format(variable_string, key_string)
            )

    return variable_keys


def is_variable_key(s):
    """
    Returns True if s is a variable string.

    :param s: The string to analyse
    :return: True, if s is a starts with VARIABLE_SEPARATOR_START and ends with VARIABLE_SEPARATOR_END
    """
    return s.startswith(VARIABLE_SEPARATOR_START) and s.endswith(VARIABLE_SEPARATOR_END)


def complete_variables(data, variables, key_string=None):
    """
    Fills the given variables into the given data.

    :param data: The data to complete
    :param variables: The variables to use
    :param key_string: A string representing the keys above the current data element.
    """
    if isinstance(data, dict):
        for key, value in data.items():
            sub_key_string = get_dict_sub_key_string(key, key_string)

            if isinstance(value, str):
                completed_variable_string = _resolve_variable_string(value, variables, sub_key_string)
                data[key] = completed_variable_string
            else:
                complete_variables(value, variables, sub_key_string)
    elif isinstance(data, list):
        for index, value in enumerate(data):
            sub_key_string = get_list_sub_key_string(index, key_string)

            if isinstance(value, str):
                completed_variable_string = _resolve_variable_string(value, variables, sub_key_string)
                data[index] = completed_variable_string
            else:
                complete_variables(value, variables, sub_key_string)


def _resolve_variable_string(variable_string, variables, key_string):
    """
    Replaces the variable keys inside the given variable string.

    :param variable_string: The string in which variables are to be replaced.
    :param variables: The variables to use
    :param key_string: The key string describing where the variable string is found in the red file.
    :return: The variable string with variable keys resolved
    """
    try:
        parts = split_into_parts(variable_string, VARIABLE_SEPARATOR_START, VARIABLE_SEPARATOR_END)
    except ParsingError as e:
        raise RedVariableError(
            'Could not parse variable "{}" in "{}". Failed with the following message:\n{}'
            .format(variable_string, key_string, str(e))
        )
    result = []
    for p in parts:
        if is_variable_key(p):
            resolved = variables[p[2:-2]]
            result.append(resolved)
        else:
            result.append(p)

    return ''.join(result)


class RedVariableError(Exception):
    pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['red_val', 'red_val.schemas']

package_data = \
{'': ['*']}

install_requires = \
['jsonschema>=4.12,<5.0']

setup_kwargs = {
    'name': 'red-val',
    'version': '9.1.1',
    'description': 'RED-val is part of the Curious Containers project. It provides functionality to implement red clients and red execution engines.',
    'long_description': '## RED-val\n\nRED-val (Reproducible Experiment Description Validation) is part of the Curious Containers project.\nIt provides functionality to implement RED-Clients namely RED-file-validation and RED-variables-resolution.\n\n## Acknowledgements\n\nThe Curious Containers software is developed at [CBMI](https://cbmi.htw-berlin.de/) (HTW Berlin - University of Applied Sciences). The work is supported by the German Federal Ministry of Economic Affairs and Energy (ZIM project BeCRF, grant number KF3470401BZ4), the German Federal Ministry of Education and Research (project deep.TEACHING, grant number 01IS17056 and project deep.HEALTH, grant number 13FH770IX6) and HTW Berlin Booster.\n',
    'author': 'Christoph Jansen',
    'author_email': 'Christoph.Jansen@htw-berlin.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.curious-containers.cc/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



engine_schema = {
    'type': 'object',
    'properties': {
        'engine': {'type': 'string'},
        'settings': {'type': 'object'},
        'doc': {'type': 'string'}
    },
    'additionalProperties': False,
    'required': ['engine', 'settings']
}



from jsonschema import ValidationError


def format_validation_error(ex, message=None):
    """
    Returns a string describing the given jsonschema.ValidationError

    :param ex: A jsonschema.ValidationError
    :type ex: ValidationError
    :param message: A message describing the error
    :type message: str
    :rtype: str
    """
    if message is None:
        message = ''

    where = '/'.join([str(s) for s in ex.absolute_path]) if ex.absolute_path else '/'
    return '{}\n\tkey in red file: {}\n\treason: {}'.format(message, where, ex.message)


class RedSpecificationError(Exception):
    pass


class RedValidationError(Exception):
    pass


class CWLSpecificationError(Exception):
    pass
